require 'uu/os/qos/qos_handler'
require 'uu/os/rest'
require 'uu/os/uesuri'
require 'uu/finman/financialtransaction/expost_payment_create'
require 'uu/finman/financialtransaction/expost_payment_attributes'
require 'uu/finman/financialtransaction/expost_payment_set_attributes'

# Module Unicorn Universe
module UU
  # Module Finman
  module Finman
    # Module of finman Financial transaction
    module FinancialTransaction
      #Module of Expost Payment
      module ExpostPayment
        extend self

        # Service path
        PATH = 'uu/finman/financialtransaction/FinancialTransaction'

        # Creates a new Expost Payment. The command creates a new Expost Payment for specified business case
        # with passed DTO parameters.<br>
        #
        # <b>Required attributes:</b><br>
        #   control_business_case_uri, counterparty_business_case_uri, name, amount_without_vat, vat_rate or amount_with_vat, financial_transaction_type, is_transaction_credit,
        #   credit_vat_date, debit_vat_date, pay_date, transaction_currency, (credit_account_number and credit_side_currrency) or credit_account_uri,
        #   (debit_account_number and debit_side_currency) or debit_account_uri
        # Exactly one of attributes <code>credit_account_number</code> and <code>credit_account_uri</code> has to be set.<br>
        # Exactly one of attributes <code>debit_account_number</code> and <code>debit_account_uri</code> has to be set.<br>
        # Exactly one of attributes <code>vat_rate</code> and <code>amount_with_vat</code> has to be set. Both attributes can be set simultaneously only if following statement is true : amount_without_vat * vat_rate = amount_with_rate.<br>
        # Attribute <code>credit_side_currency</code> and <code>debit_side_currency</code> must be set if respective account is set by number.
        # Attribute <code>transaction_currency</code> must be set and must be one from <code>credit_side_currency</code> and <code>debit_side_currency</code>, with exception if both <code>credit_side_currency</code> and <code>debit_side_currency</code> are equal.
        #
        # @param [UU::OS::UESURI] control_business_case_uri UESURI of the business case of the control side to create Expost Payment for.
        # @param [UU::Finman::FinancialTransaction::ExpostPaymentCreate] expost_payment DTO containing attributes of the new Expost Payment
        #
        # @return [UU::OS::UESURI] UESURI of the new Expost Payment. In case of setting <code>dry_run</code> parameter to <code>true</code>, <code>UU::OS::UESURI::NIL_URI</code> is returned instead.
        #
        #
        # @example
        #      # Prepare Business Case URI, which we want to use to create Expost payment.
        #      locationUri = UU::OS::UESURI.new('ues:TERRITORY:BUSINESS_CASE');
        #
        #      # Create Expost payment
        #        UU::Finman::FinancialTransaction::ExpostPayment.create(control_business_case_uri,
        #        :name => "Ex-post payment ",
        #        :credit_description => "kreditní popis",
        #        :debit_description => "debetní popis",
        #        :credit_financial_category => UU::OS::UESURI.new("ues:TERRITORY:APPS.UU.FINMAN/CONFIGURATION:SYS.FIN.CONFIG/TX_TYPE/CR.TXWER"),
        #        :debit_financial_category => UU::OS::UESURI.new("ues:TERRITORY:APPS.UU.FINMAN/CONFIGURATION:SYS.FIN.CONFIG/TX_TYPE/DR.TXOTR"),
        #        :is_transaction_credit => true,
        #        :financial_transaction_type => "NONCASH",
        #        :amount_without_vat => 10000,
        #        :vat_rate => 10,
        #        :invoice_number => "Číslo faktury",
        #        :constant_symbol => "konstantní symbol",
        #        :variable_symbol => "variabilní symbol",
        #        :specific_symbol => "specifický symbol",
        #        :credit_vat_date => "2014-11-03",
        #        :debit_vat_date => "2014-11-04",
        #        :pay_date => "2014-10-06",
        #        :counterparty_business_case_uri => UU::OS::UESURI.new("ues:TERRITORY:CP_BC"),
        #        :credit_account_number => "kreditní číslo učtu",
        #        :debit_account_number => "debetní číslo učtu",
        #        :is_planned => false,
        #        :contract_uri => UU::OS::UESURI.new("ues:TERRITORY:CONTRACT"),
        #        :credit_side_currency => "CZK",
        #        :debit_side_currency => "EUR",
        #        :transaction_currency => "EUR"
        #       )
        def create(control_business_case_uri, expost_payment)
          svc = UU::OS::REST::RemoteClient.new(ExpostPayment)
          payment_dto = UU::Finman::FinancialTransaction::ExpostPaymentCreate.new(expost_payment)
          # adjust given Expost Payment dto to match expected format of financial transaction command
          transaction = transform_expost_payment_create_to_transaction_dto(payment_dto)
          financial_transaction_create = transaction.to_json

          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.post('create', control_business_case_uri, financial_transaction_create)
            if res.nil?
              return res
            else
              return UU::OS::UESURI.new(res)
            end
          end
        end

        # Obtains Ex-post attributes from service and returns them
        #
        # @param [String, UU::OS::UESURI] expost_payment_uri UESURI of the Expost Payment
        #
        # @return [ExpostPaymentAttributes] ExpostPayment attributes
        #
        #
        # @example
        #      # Prepare Expost payment location which we want get attributes.
        #      expost_payment_uri = UU::OS::UESURI.new('ues:TERRITORY:EXPOST_PAYMENT');
        #
        #      # Get Expost payment
        #      UU::Finman::FinancialTransaction::ExpostPayment.get_attributes(expost_payment_uri);
        def get_attributes(expost_payment_uri)
          svc = UU::OS::REST::RemoteClient.new(ExpostPayment)
          UU::OS::QoS::QoSHandler.auto_retry do
            attrs_string = svc.get('getAttributes', expost_payment_uri)
            attrs = JSON.parse(attrs_string, :symbolize_names => true)
            attrs = remove_invalid_keys(attrs)
            return UU::Finman::FinancialTransaction::ExpostPaymentAttributes.new(attrs)
          end
        end

        # Sets attributes of Expost Payment.
        #
        # @param [String,UU::OS::UESURI] expost_payment_uri UESURI of modified Payment
        # @param [ExpostPaymentSetAttributes] expost_payment_set_attributes DTO containing modified attributes of the Expost Payment
        #
        # @return [UU::OS::UESURI] UESURI of modified Payment
        #
        #
        # @example
        #      # Prepare Expost payment location which we want set attributes.
        #      expost_payment_uri = UU::OS::UESURI.new('ues:TERRITORY:EXPOST_PAYMENT');
        #
        #      # Create Expost payment
        #      UU::Finman::FinancialTransaction::ExpostPayment.set_attributes(expost_payment_uri,
        #        :name => "Expost payment set");
        def set_attributes(expost_payment_uri, expost_payment_set_attributes)
          svc = UU::OS::REST::RemoteClient.new(ExpostPayment)
          payment_dto = UU::Finman::FinancialTransaction::ExpostPaymentSetAttributes.new(expost_payment_set_attributes)
          # adjust given payment dto to match expected format of financial transaction command
          transaction = transform_expost_payment_modify_to_transaction_dto(payment_dto)
          financial_transaction_dto = transaction.to_json

          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.post('setAttributes', expost_payment_uri, financial_transaction_dto)
            return UU::OS::UESURI.new(res)
          end
        end



        # Adjust dto of payment to match requirements of financial transaction dto of server side.
        # @private
        def transform_expost_payment_create_to_transaction_dto(payment_dto)
          if payment_dto.is_a?(UU::OS::REST::DTO)
            attributes = payment_dto.to_hash

            # set class of transaction to determine Payment at the server side.
            attributes[:financialTransactionClass] = "EXPOST_PAYMENT"
            #expost payments always have 100% probability
            attributes[:realizationProbability] = 100

            # remove unwanted attributes (in case that e.g. hashmap with attributes from other FinTx was modified and used)
            attributes.delete(:creditBusinessCaseFromOppositeTerritory)
            attributes.delete(:debitBusinessCaseFromOppositeTerritory)
            attributes.delete(:metaArtifactUri)
            attributes.delete(:requestApprovalByManager)
            attributes.delete(:waitForAcceptanceByAccountant)
            attributes.delete(:dueDate)
            attributes.delete(:expectedPayDate)
            attributes.delete(:isPlanned)
            attributes.delete(:isReadyForAcceptance)
            attributes.delete(:forAcceptanceDate)
            attributes.delete(:protocolDate)
            return attributes
          end
          fail "Given object is neither UU::OS::REST::DTO nor its descendant!"
        end
        private :transform_expost_payment_create_to_transaction_dto

        # Hash returned from service contains all attributes from financial transaction.
        # Remove invalid keys to contain only values for this specific class of transaction.
        # @private
        def remove_invalid_keys(hash)
          if hash.is_a?(Hash)
            hash.delete(:acceptedDate)
            hash.delete(:complementaryTransferUri)
            hash.delete(:creditBusinessCaseFromOppositeTerritory)
            hash.delete(:debitBusinessCaseFromOppositeTerritory)
            hash.delete(:dueDate)
            hash.delete(:expectedPayDate)
            hash.delete(:forAcceptanceDate)
            hash.delete(:interterritorialPaymentCopyUri)
            hash.delete(:interterritorialPaymentOriginalUri)
            hash.delete(:isPlanned)
            hash.delete(:isReadyForAcceptance)
            hash.delete(:protocolDate)
            hash.delete(:reinvoiceFromCategoryUri)
            hash.delete(:reinvoiceToCategoryUri)
            hash.delete(:reinvoiceFromItem)
            hash.delete(:reinvoiceOnItem)
            hash.delete(:reinvoicingType)
            hash.delete(:requestApprovalByManager)
            hash.delete(:waitForAcceptanceByAccountant)
            hash.delete(:closeAfterCreation)
            return hash
          end
          fail "Given object must be Hash!"
        end
        private :remove_invalid_keys

        # Adjust dto of payment to match requirements of financial transaction dto of server side.
        # @private
        def transform_expost_payment_modify_to_transaction_dto(payment_dto)
          if payment_dto.is_a?(UU::OS::REST::DTO)
            attributes = payment_dto.to_hash
            # set class of transaction to determine Payment at the server side.
            attributes[:financialTransactionClass] = "EXPOST_PAYMENT"
            # remove unwanted attributes (in case that e.g. hashmap with attributes from other FinTx was modified and used)
            attributes.delete(:creditBusinessCaseFromOppositeTerritory)
            attributes.delete(:debitBusinessCaseFromOppositeTerritory)
            attributes.delete(:requestApprovalByManager)
            attributes.delete(:waitForAcceptanceByAccountant)
            attributes.delete(:dueDate)
            attributes.delete(:expectedPayDate)
            attributes.delete(:isPlanned)
            attributes.delete(:forAcceptanceDate)
            attributes.delete(:protocolDate)
            return attributes
          end
          fail "Given object is neither UU::OS::REST::DTO nor its descendant!"
        end
        private :transform_expost_payment_modify_to_transaction_dto

      end
    end
  end
end
